/* pm1dump.c -- Copyright (C) 2004, Geoffrey Reynolds.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   The pm1dump program reads a P-1 save file created by Prime95/mprime
   and writes a save file on standard output in a format suitable for
   continuing P-1 with gmp-ecm using the -resume option. The number
   being tested is divided through by any known factors in low[mp].txt
   so that gmp-ecm works with just the remaining composite cofactor.

   For large exponents mprime completes stage one P-1 testing faster
   than gmp-ecm, especially so for Pentium 4 machines. However gmp-ecm
   has a better stage two algorithm, and can perform both stage one
   and two to deeper bounds than mprime. Thus it is often useful to do
   stage one using mprime then continue with gmp-ecm. 

   Here is an example session for P-1 on 2^1073-1 using mprime to do
   stage one to bound B1=300000 and gmp-ecm to continue stage two to
   bound B2=30000000000 (long lines have been split):

   $ 
   $ cat worktodo.ini
   Pminus1=1073,3000000,1,0,0
   $ 
   $ cat lowm.txt 
   M( 1073 )C: 223
   M( 1073 )C: 233
   M( 1073 )C: 1103
   M( 1073 )C: 2089
   M( 1073 )C: 616318177
   $ 
   $ mprime -d
   Mersenne number primality test program version 23.5
   P-1 on M1073 with B1=3000000, B2=3000000
   M1073 stage 1 complete. 11381642 transforms. Time: 6.059 sec.
   Stage 1 GCD complete. Time: 0.001 sec.
   There are no more exponents to test.
   Please send the results.txt file to woltman@alum.mit.edu
   Contact the PrimeNet server for more exponents.
   $ 
   $ pm1dump m0001073 | ecm -resume - 3e6 3e10
   GMP-ECM 5.0.3 [powered by GMP 4.1.2] [ECM]
   Resuming P-1 residue 
   Input number is 13715310197630769733432134564056731689035607493321133674 \
   772367199503243546539527407434627283722858590475006767982474856221525534 \
   239707683043256747375324173635170964119563015203960914200228737031313766 \
   708508964922446191966703478484449296915471876339736972327863891284991160 \
   90302096958459493737084078784511 (304 digits)
   Using B1=3000000-3000000, B2=30000000000, polynomial x^60
   Step 1 took 0ms
   Step 2 took 396782ms
   ********** Factor found in step 2: 44197213979727234355841
   Found probable prime factor of 23 digits: 44197213979727234355841
   Composite cofactor 31032069586833750370109936900385095506462082516159002 \
   331799590091443688163779085607869137785636692987202986504147255894061331 \
   130150720345307481198824606984551750514880702222748081599256700869626491 \
   456490483111490127368069661821121997479502005893372259937838386976970831 \
   974873567871 has 281 digits
   $ 


   I have tested this program with mprime v23.5 and gmp-ecm v5.0.3,
   but only on a limited range of possible input. If you find a bug
   please email me at geoff@hisplace.co.nz.

   This code uses the gmp library and some GNU extensions to the
   standard C library. */

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <gmp.h>

#define PROGRAM_NAME "pm1dump"
#define PROGRAM_VERSION "0.3"

#define PM1_MAGIC 0x1a2b3c4d

#define LOWM_TXT_DEFAULT "lowm.txt"
#define LOWP_TXT_DEFAULT "lowp.txt"

static void usage(void)
{
  printf("Usage: %s [OPTIONS] [INFILE ...] > OUTFILE\n", PROGRAM_NAME);
  printf("Options:\n");
  printf("-m FILE  Look for known factors of 2^N-1 in FILE instead of %s.\n",
         LOWM_TXT_DEFAULT);
  printf("-p FILE  Look for known factors of 2^N+1 in FILE instead of %s.\n",
         LOWP_TXT_DEFAULT);
  printf("-c       Insert a comment describing the output.\n");
  printf("-h       Print this message.\n");
  printf("-V       Print the program version number.\n");
  printf("\n");
}

static void version(void)
{
  printf("%s version %s\n", PROGRAM_NAME, PROGRAM_VERSION);
}

/* Print the number being P-1 tested after dividing out known factors */
void print_N(unsigned long exponent, int plus1, const char *low_txt, int cflag)
{
  unsigned long n;
  char c;
  FILE *file;
  mpz_t F, N;

  if (low_txt != NULL) /* file name supplied on command line */
  {
    if ((file = fopen(low_txt, "r")) == NULL)
      fprintf(stderr, "%s: cannot open %s.\n", PROGRAM_NAME, low_txt);
  }
  else /* use default, don't complain if missing */
  {
    low_txt = plus1 ? LOWP_TXT_DEFAULT : LOWM_TXT_DEFAULT;
    file = fopen(low_txt, "r");
  }

  mpz_init(N);
  mpz_ui_pow_ui(N, 2, exponent);
  if (plus1)
    mpz_add_ui(N, N, 1);
  else
    mpz_sub_ui(N, N, 1);

  if (file != NULL)
  {
    size_t len = 0;
    char *buf = NULL;
#ifdef DEBUG
    int started = 0;
#endif

    mpz_init(F);

    /* Divide through by factors in low_txt. Silently ignore any line
       that does not parse. */
    while (getline(&buf, &len, file) != EOF)
    {
      if (gmp_sscanf(buf, "%c( %lu )C: %Zu\n", &c, &n, F) == 3)
      {
        if (exponent == n && ((plus1 && c == 'P') || (!plus1 && c == 'M')))
        {
          if (mpz_remove(N, N, F) == 0)
            gmp_fprintf(stderr, "%s: %s %s 2^%lu%c1: %Zu.\n", PROGRAM_NAME,
                        low_txt, "contains bad or duplicate factor for",
                        exponent, plus1? '+' : '-', F);
#ifdef DEBUG
          else if (cflag)
          {
            if (started++ == 0)
              printf("# %s, Factors in %s:", PROGRAM_NAME, low_txt);
            gmp_printf(" %Zu", F);
          }
#endif
        }
      }
    }
#ifdef DEBUG
    if (started)
      printf(".\n");
#endif
    if (buf != NULL)
      free(buf);
    mpz_clear(F);
    fclose(file);
  }

  gmp_printf("N=0x%ZX;", N);
  mpz_clear(N);
  return;
}

/* This function is an inverse of tempFileName() in source/commonc.c. */
int parse_filename(const char *name, unsigned long *exponent, int *plus1)
{
  unsigned long p = 0, q = 0;
  int i;

  if (strlen(name) != 8)
    return 0;

  switch (name[0])
  {
    case 'l': *plus1 = 1; break;
    case 'm': *plus1 = 0; break;
    default: return 0;
  }

  for (i = 1; i <= 7; i++)
  {
    if (isdigit(name[i]))
      p = p * 10 + (name[i] - '0');
    else
    {
      switch (i)
      {
        case 1:
          if (name[i] < 'A' || name[i] > 'A' + 24)
            return 0;
          q = 10000000;
          break;
        case 2:
          if (name[i] < 'A' || name[i] > 'A' + 34 || q > 0)
            return 0;
          q = 35000000;
          break;
        case 3:
          if (name[i] < 'A' + 10 || name[i] > 'A' + 19 || q > 0)
            return 0;
          q = 60000000;
          break;
        default:
          return 0;
      }

      q += 1000000 * (name[i] - 'A');
    }
  }

  *exponent = p + q;
  return *exponent != 0;
}

int read_save_file(const char *name, unsigned long *B1, long *len, long **buf)
{
  FILE *file;
  unsigned long tmp, version;

  int file_error(const char *msg)
  {
    fprintf(stderr, "%s: %s: %s.\n", PROGRAM_NAME, name, msg);
    if (file)
      fclose(file);
    return 0;
  }

  if ((file = fopen(name, "r")) == NULL)
    return file_error("cannot open file");

  if (fread(&tmp, sizeof(unsigned long), 1, file) != 1)
    return file_error("error reading from file");

  if (tmp != PM1_MAGIC)
    return file_error("not a Prime95 P-1 save file");

  if (fread(&version, sizeof(unsigned long), 1, file) != 1)
    return file_error("error reading from file");

  if (version < 3 || version > 4)
    return file_error("not a version 3-4 save file");

  if (fseek(file, sizeof(long), SEEK_CUR))
    return file_error("error seeking in file");

  if (fread(B1, sizeof(unsigned long), 1, file) != 1)
    return file_error("error reading from file");

  if (version == 3)
  {
    if (fseek(file, 7 * sizeof(long), SEEK_CUR))
      return file_error("error seeking in file");
  }
  else if (version == 4)
  {
    if (fseek(file, 8 * sizeof(long), SEEK_CUR))
      return file_error("error seeking in file");
    if (fread(&tmp, sizeof(unsigned long), 1, file) != 1)
      return file_error("error reading from file");
    if (fseek(file, tmp*sizeof(char) + 2*sizeof(long), SEEK_CUR))
      return file_error("error seeking in file");
  }

  if (fread(len, sizeof(long), 1, file) != 1)
    return file_error("error reading residue length");

  if (*len <= 0)
    return file_error("invalid residue length");

  if ((*buf = malloc((*len)*sizeof(long))) == NULL)
    file_error("malloc failed");

  if (fread(*buf, (*len)*sizeof(long), 1, file) != 1)
  {
    free(buf);
    return file_error("error reading residue data");
  }

  fclose(file);
  return 1;
}

int main(int argc, char **argv)
{
  int c, i, plus1, error = 0, cflag = 0;
  unsigned long exponent, B1;
  long len = 0, *buf = NULL;
  const char *mval = NULL, *pval = NULL;

  while ((c = getopt(argc, argv, "m:p:chV")) != -1)
    switch (c)
    {
      case 'm': mval = optarg; break;
      case 'p': pval = optarg; break;
      case 'c': cflag = 1; break;
      case 'h': usage(); break;
      case 'V': version(); break;
      default: error = 1;
    }

  for (c = optind; c < argc; c++)
  {
    if (!parse_filename(basename(argv[c]), &exponent, &plus1))
    {
      fprintf(stderr, "%s: %s is not the name of a Prime95 P-1 save file.\n",
              PROGRAM_NAME, argv[c]);
      error = 1;
      continue;
    }

    if (read_save_file(argv[c], &B1, &len, &buf))
    {
      if (cflag)
        printf("# %s, %s is a Prime95 P-1 save file for 2^%lu%c1:\n",
               PROGRAM_NAME, argv[c], exponent, plus1 ? '+' : '-');

      print_N(exponent, plus1, plus1 ? pval : mval, cflag);

      /* Print the stage one residue, a giantint. */
      printf(" X=0x%lX", buf[len-1]);
      for (i = (len - 2); i >= 0; i--)
        printf("%08lX", buf[i]);

      printf("; B1=%lu;", B1);
      printf(" METHOD=P-1;");
      printf("\n");

      free(buf);
    }
    else error = 1;
  }

  return error;
}
