/* splitcf.c -- Copyright (C) 2004, Geoffrey Reynolds.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   The splitcf program scans standard input for ECM stage one save
   records as created by mprime (when run with the GmpEcmHook=1 option
   in prime.ini). If one is found for a divisor N of 2^(2n)-1 then
   new records are created for the factors of N which divide 2^n-1
   (recursively if n is even) and 2^n+1 (or the Aurifeullian factors
   of 2^n+1 if n=4k-2).

   For example if N divides 2^3004-1 then the recors is split into
   separate records for each of the non-trivial dividors of N that
   divide 2,751-; 2,751+; 2,1502L; and 2,1502M respectively.

   Because the time mprime takes to complete stage one ECM for 2^2n-1
   is much less than the time to complete the same curve for 2^n-1 and
   2^n+1 separately, while stage two ECM with gmp-ecm may be the other
   way around, it is often desirable to do stage one for 2^2n-1 with
   mprime followed by stage two for 2^n-1 and 2^n+1 with gmp-ecm.

   Also when working on the Aurifeullian factors of 2^n+1 (when n=4k-2)
   it is often best to do stage one on 2^n+1 with mprime and then
   stage two on the separate factors 2,nL and 2,nM with gmp-ecm.

   splitcf simplifies the procedure for doing this from the command
   line. For example to run 100 curves to search for 50 digit factors
   of 2^859-1, 2^859+1 and 2^1718+1 using this method, add the
   following line to mprime's worktodo.ini

     ECM=3436,43000000,1,100,0,0,0,0

   and run mprime and gmp-ecm with

     $ mprime; splitcf < results.txt | ecm -resume - 1 43e6-18e10

   This actually runs 400 separate curves, 100 each for 2^859-1,
   2^859+1, and the Aurifeullian factors 2,1718L and 2,1718M of
   2^1718+1. For this to work properly all the known factors of
   2^3436-1 must be added to lowm.txt before running mprime.


   The following command line options are accepted by splitcf:

   -c  Insert a description line before each split subfactor.
   -p  Let any junk text in the input pass through unaltered.
   -h  Print this text.
   -V  Print the program version number.

   The -c option produces a comment line before the split subfactor
   which contains the number (in Cuningham project notation) that the
   subfactor divides. This is useful in combination with grep to
   collect all the subfactors of the same number into one file.

   For example to run 100 curves to search for 40 digit factors of
   2,1598L (the Aurifeullian L factor of 2^1598+1) using mprime for
   stage one, add this line to worktodo.ini

     ECM=3196,3000000,1,100,0,0,0,0  (for sse2 machines)

   or

     ECM=1598,3000000,1,100,0,0,1,0  (for non-sse2 machines)

   and then run mprime and gmp-ecm with

     $ mprime
     $ splitcf -c < results.txt | grep -A1 "2,1598L" | ecm -resume - 1 3e6-4e9

   Again, the known factors of 2^3196-1 must be in lowm.txt (or the
   known factors of 2^1598+1 in lowp.txt for the non-sse2 case).


   I have tested this program with mprime v23.9 and gmp-ecm v5.0.3,
   but only on a limited range of possible input. If you find a bug
   please email me at geoff@hisplace.co.nz.


   This code uses the gmp library (which anyone running gmp-ecm will
   have) and the GNU getline and getopt extensions to the standard C
   library. It also uses some gcc extensions to the C language. All
   of these should be available on standard Linux, FreeBSD, or Cygwin
   systems I think. */


#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <gmp.h>

#define PROGRAM_NAME "splitcf"
#define PROGRAM_VERSION "0.3"

/* Stop checking whether the input composite is a factor of 2^n-1 if n
   reaches this value. */
#define MAX_N 79300000

static void usage(void)
{
  printf("Usage: %s [OPTIONS] < INFILE > OUTFILE\n", PROGRAM_NAME);
  printf("-c  Insert a description line before each split subfactor.\n");
  printf("-p  Let any junk text in the input pass through unaltered.\n");
  printf("-h  Print this text.\n");
  printf("-V  Print the program version number.\n");
  printf("\n");
}

static void version(void)
{
  printf("%s version %s\n", PROGRAM_NAME, PROGRAM_VERSION);
}

int main(int argc, char **argv)
{
  char qx[3], *buf = NULL;
  size_t buflen = 0, buftail;
  unsigned n;
  int ch, cflag = 0, pflag = 0;
  mpz_t N, M, A1, A2, R, R1;

  void print_rec(mpz_t N, char c)
  {
    if (mpz_cmp_ui(N,1) != 0)
    {
      if (cflag)
        printf("# %s, factor of 2,%d%c:\n", PROGRAM_NAME, n, c);
      mpz_mod(R1,R,N); /* reduce the input residue modulo N */
      gmp_printf("N=0x%ZX; %s=0x%ZX;%s", N, qx, R1, buf+buftail);
    }
  }

  void splitcf(void)
  {
    if (n % 2 == 1)
    {
      print_rec(N,'-');
      return;
    }

    n = n / 2;
    if (n % 4 == 2)  /* so that n = 2h, where h=2k-1 */
    {
      unsigned h = n / 2;
      unsigned k = (h + 1) / 2;

      /* Calculate Aurifeullian L factor */
      mpz_ui_pow_ui(A1,2,h);    /* A1 = 2^h   */
      mpz_add_ui(A1,A1,1);      /*    = 2^h+1 */
      mpz_ui_pow_ui(A2,2,k);    /* A2 = 2^k   */
      mpz_sub(M,A1,A2);         /* M  = 2^h-2^k+1 */
      mpz_gcd(M,N,M);
      print_rec(M,'L');

      /* Calculate Aurifeullian M factor */
      mpz_add(M,A1,A2);         /* M  = 2^h+2^k+1 */
      mpz_gcd(M,N,M);
      print_rec(M,'M');
    }
    else
    {
      mpz_ui_pow_ui(M,2,n);     /* M = 2^n */
      mpz_add_ui(M,M,1);        /*   = 2^n+1 */
      mpz_gcd(M,N,M);
      print_rec(M,'+');
    }

    if (n > 1)
    {
      mpz_ui_pow_ui(M,2,n);     /* M = 2^n */
      mpz_sub_ui(M,M,1);        /*   = 2^n-1 */
      mpz_gcd(N,N,M);
      if (mpz_cmp_ui(N,1) != 0)
        splitcf();
    }

    return;
  }

  /* Begin main() */

  while ((ch = getopt(argc, argv, "cphV")) != -1)
    switch (ch)
    {
      case 'c': cflag = 1; break;
      case 'p': pflag = 1; break;
      case 'h': usage(); return 0;
      case 'V': version(); return 0;
      default:  return 1;
    }

  mpz_init(N);  /* input composite */
  mpz_init(M);  /* general temporary */
  mpz_init(A1); /* Aurifeullian intermediate */
  mpz_init(A2); /* Aurifeullian intermediate */
  mpz_init(R);  /* input residue */
  mpz_init(R1); /* reduced residue temporary*/

  while (getline(&buf, &buflen, stdin) != EOF)
  {
    if (gmp_sscanf(buf, "N=0x%ZX; %2[QX]=0x%ZX;%n", N, qx, R, &buftail) == 3)
    {
      n = mpz_sizeinbase(N,2);
      mpz_ui_pow_ui(M,2,n-1);
      do /* find the smallest n <= MAX_N such that N divides 2^n-1 */
      {
        mpz_mul_ui(M,M,2);
        mpz_mod(M,M,N);
      }
      while (mpz_cmp_ui(M,1) != 0 && ++n < MAX_N);

      if (n < MAX_N)
        splitcf();          /* Split and print this record */
      else
        printf("%s", buf);  /* Print the unsplit record.*/
      continue;
    }

    if (pflag)
      printf("%s", buf);    /* Let any junk lines pass through. */
  }

  return 0;
}
